/* $Id: ld.c,v 1.5 2000/11/27 20:09:44 malekith Exp $ */

#include "util.h"
#include "binfmt.h"
#include "op.h"
#include "phash.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

byte startup[] = { 
	op_nop, 
	op_nop, 
	op_nop, 
	op_nop,
	op_nop, 
	op_nop, 
	op_nop, 
	/* align to 8, should be enough for anything ... */
	/* call Sys::Sys() */
	op_new, 0, 0, 0, 0, 
	/* we have sys_ptr in %a */
	/* now call Sys::die() */
	op_mov_b, 1,
	op_jsr,
	/* we should not get so far, but ... */
	op_mov_b, 0xff,
	op_push_d,
	op_clr_d,
	/* call exit(-1) */
	op_syscall,
	/* ??? */
	op_hlt, 
	0, 0, 0
};
/* gen:
one text, entry at 0
vtables:
	u32 nclasses;
	{u32 nfunc, func[nfuncs]}[nclasses];
*/

struct file {
	i32 nclasses;
	struct class_def **classes;
	const char *name;

	byte *text;
	int text_base;

	byte *class_xrefs;
	int class_xrefs_size;

	byte *funcs_xrefs;
	int funcs_xrefs_size;

	byte *funcs;
	int funcs_size;
};

struct class_def {
	struct class_def *left, *right;
	char *name;
	int id;
	
	int nfuncs;
	u32 *def_funcs;
	struct class_def **func_xrefs;
};

struct class_def *root;
struct dbuf *text;

static struct class_def *new_class_def(char *name)
{
	struct class_def *c;
	static int id = 1;

	c = NEW(struct class_def);
	memset(c, 0, sizeof(*c));
	c->name = strdup(name);
	if (strcmp(c->name, "Sys") == 0)
		c->id = 0;
	else
		c->id = id++;

	return c;
}

static int parse_classes(struct file *f, struct kbin_tag *t)
{
	char *p;
	int i, r;
	struct class_def *c, *n;
	
	f->nclasses = *(u32*)t->data;
	p = (char*)(t->data + sizeof(u32));
	f->classes = (struct class_def**)malloc(
			f->nclasses * sizeof(struct class_def*));
			
	for (i = 0; i < f->nclasses; i++) {
		if (root == 0)
			root = n = c = new_class_def(p);
		else
			for (n = c = root; n; c = n) {
				r = strcmp(c->name, p);
				if (r < 0)
					n = c->left;
				else if (r > 0)
					n = c->right;
				else
					break;
				if (n == 0)
					break;
			}
		if (n == 0) {
			n = new_class_def(p);
			r = strcmp(n->name, p);
			if (r < 0)
				c->left = n;
			else
				c->right = n;
		}
		
		f->classes[i] = n;
		
		p = strchr(p, 0);
		p++;
		if ((byte*)p - t->data > t->len || 
		    ((byte*)p - t->data == t->len && i + 1 < f->nclasses)) {
		    	fprintf(stderr, "%s: size error in class list\n",
		    		f->name);
		    	return -1;
		}
	}
	
	if ((byte*)p - t->data != t->len) {
		fprintf(stderr, "%s: size+ error in class list\n",
			f->name);
		return -1;
	}

	return 0;
}

/*
 * map local to global ids in text
 */
static int resolve_class_xrefs(struct file *f)
{
	u32 *ptr;
	int n;
	int id;

	n = f->class_xrefs_size / sizeof(u32);
	
	for (ptr = (u32*)f->class_xrefs; n--; ptr++) {
		id = *(u32*)(f->text + *ptr);
		if (id >= f->nclasses) {
			fprintf(stderr, "%s: bad id\n", f->name);
			return -1;
		}
		*(u32*)(f->text + *ptr) = f->classes[id]->id;
	}
	free(f->class_xrefs);
	f->class_xrefs = 0;

	return 0;
}

/*
 * map table describing where (in which class) to find given function
 * from local to class_def*'s
 */
static int resolve_funcs_xrefs(struct file *f)
{
	i32 *ptr;
	int nfn, i, j;
	struct class_def *c;

	ptr = (i32*)f->funcs_xrefs;
	
	if (*ptr++ != f->nclasses) {
		fprintf(stderr, "%s: number of classes not consistent\n",
			f->name);
		return -1;
	}
	
	for (i = 0; i < f->nclasses; i++) {
		nfn = *ptr++;
		c = f->classes[i];
		if (c->def_funcs) {
			if (c->nfuncs != nfn) {
				fprintf(stderr, "%s: no of func changed "
					"in class %s", f->name, c->name);
				return -1;
			}
			/* FIXME: compare */
			ptr += nfn;
		} else {
			c->nfuncs = nfn;
			c->def_funcs = (u32*)malloc(nfn * sizeof(u32));
			memset(c->def_funcs, 0, nfn * sizeof(u32));
			c->func_xrefs = (struct class_def**)malloc(
					nfn * sizeof(struct class_def*));
			for (j = 0; j < nfn; j++) {
				if (*ptr >= f->nclasses) {
					fprintf(stderr, "%s: bad funcid\n",
						f->name);
					return -1;
				}
				c->func_xrefs[j] = f->classes[*ptr];
				ptr++;
			}
		}
	}
	free(f->funcs_xrefs);
	f->funcs_xrefs = 0;

	return 0;
}

/*
 * put all actual definition of functions in given file into
 * appropiate places in class_def tables
 */
static int resolve_funcs(struct file *f)
{
	i32 *ptr;
	int n;
	struct class_def *c;
	
	n = f->funcs_size / sizeof(u32);
	
	for (ptr = (i32*)f->funcs; n-- > 0; ptr++) {
		if (*ptr >= f->nclasses) {
			fprintf(stderr, "%s: bad func id\n", f->name);
			return -1;
		}
		c = f->classes[*ptr++];
		n--;
		if (*ptr >= c->nfuncs) {
			fprintf(stderr, "%s: too big func definied\n", 
				f->name);
			return -1;
		}
		if (c->func_xrefs[*ptr] != c) {
			fprintf(stderr, "%s: this function shall not be "
				"definied in its own class\n", 
				f->name);
			return -1;
		}
		c->def_funcs[*ptr] = ptr[1] + f->text_base;
		ptr++;
		n--;
	}
	
	free(f->funcs);
	f->funcs = 0;

	return 0;
}

int parse_object_file(const char *fn)
{
	struct file f;
	struct kbin_file *o;
	struct kbin_tag *t;
	
	memset(&f, 0, sizeof(f));

	f.name = fn;
	o = kbin_open(fn, 1, 0);
	if (o == 0)
		return -1;
	while ((t = kbin_get_tag(o))) {
		if (t->id == kbin_tag_eof) {
			kbin_kill_tag(t);
			break;
		}
		
		switch (t->id) {
		case kbin_tag_code:
			f.text_base = db_off(text);
			f.text = db_add(text, t->data, t->len);
			break;
		case kbin_tag_ok_funcs_xrefs:
			f.funcs_xrefs = t->data;
			f.funcs_xrefs_size = t->len;
			t->data = 0;
			break;
		case kbin_tag_ok_class_xrefs:
			f.class_xrefs = t->data;
			f.class_xrefs_size = t->len;
			t->data = 0;
			break;
		case kbin_tag_ok_funcs:
			f.funcs = t->data;
			f.funcs_size = t->len;
			t->data = 0;
			break;
		case kbin_tag_ok_classes:
			if (parse_classes(&f, t))
				return -1;
			break;
		}
		
		kbin_kill_tag(t);
	}
	
	kbin_close(o);
	
	if (t == 0)
		return -1;

	if (resolve_class_xrefs(&f) ||
	    resolve_funcs_xrefs(&f) ||
	    resolve_funcs(&f))
		return -1;

	return 0;
}

int check_symbols(struct class_def *c)
{
	int i, unr = 0;
	struct class_def *r;
	
	if (c == 0)
		return 0;
		
	for (i = 0; i < c->nfuncs; i++)
		if (c->func_xrefs[i] == c) {
			if (c->def_funcs[i] == 0)
				unr++;
		} else {
			r = c->func_xrefs[i];
			if (r->nfuncs <= i) {
				fprintf(stderr, "base %s of %s hasn't got "
					"enough symbols to resolve function "
					"#%d", r->name, c->name, i);
				return -1;
			}
			if (r->func_xrefs[i] != r) {
				fprintf(stderr, "symbol #%d isn't resolved "
					"directly in base %s of %s",
					i, r->name, c->name);
				return -1;
			}
		}

	if (unr)
		fprintf(stderr, "class %s has %d unresolved symbol(s)\n",
			c->name, unr);
	
	return check_symbols(c->left) || check_symbols(c->right) || unr;
}

void resolve_symbols(struct class_def *c)
{
	int i;
	
	if (c == 0)
		return; 
		
	for (i = 0; i < c->nfuncs; i++)
		if ((c->def_funcs[i] = c->func_xrefs[i]->def_funcs[i]) == 0)
			halt();

	resolve_symbols(c->left);
	resolve_symbols(c->right);
}

struct class_def *find_min_id(struct class_def *p)
{
	struct class_def *p1, *p2, *px;
	
	if (p == 0)
		return 0;
	
	p1 = find_min_id(p->left);
	p2 = find_min_id(p->right);

	if (p1 && p1->id != -1 && p1->id < p2->id)
		px = p1;
	else
		px = p2;
	
	if (px && px->id != -1 && (p->id == -1 || px->id < p->id)) 
		return px;
	else 
		return p;
}

void make_vtable(struct kbin_file *o)
{
	struct dbuf *db;
	int i = 0;
	struct class_def *c;
	
	db = db_new();
	db_alloc(db, sizeof(u32));
	for (;;) {
		c = find_min_id(root);
		if (c->id == -1)
			break;
		if (c->id != i++)
			halt();
		c->id = -1;
		*(u32*)db_alloc(db, sizeof(u32)) = c->nfuncs;
		db_add(db, c->def_funcs, sizeof(u32) * c->nfuncs);
	}
	*(u32*)db_data(db) = i;
	db = db_finish(db, &i);
	kbin_add_tag(o, kbin_tag_vtables, i, (byte*)db);
	free(db);
}

void dump_text(struct kbin_file *o)
{
	byte *b;
	int i;
	
	b = db_finish(text, &i);
	kbin_add_tag(o, kbin_tag_code, i, b);
	free(b);
}

void init_text()
{
	text = db_new();
	assert(sizeof(startup) == 24);
	db_add(text, startup, 24);
}

int main(int argc, char **argv)
{
	int i;
	struct kbin_file *f;
	
	if (argc <= 2) {
		fprintf(stderr, "USAGE: %s target objects...\n", *argv);
		exit(1);
	}

	init_text();
	
	for (i = 2; i < argc; i++)
		if (parse_object_file(argv[i]))
			break;
	if (i < argc)
		exit(1);
		
	if (check_symbols(root))
		exit(1);

	resolve_symbols(root);

	f = kbin_open(argv[1], 0, 0);
	if (f == 0)
		exit(1);
	make_vtable(f);
	dump_text(f);
	kbin_fini(f);

	return 0;
}
