/* $Id: vm.c,v 1.6 2000/11/23 00:45:42 malekith Exp $ */

#include "util.h"

#include <string.h>
#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>

#include "vm.h"
#include "ccalls.h"
#include "binfmt.h"

int brk = -1;

#define STACK_SIZE 1000

/* now, instructions are aligned with nop's */
#define align(cpu)	do { } while (0)

#define get_unaligned(dst, type)			\
	do {						\
		align(cpu);				\
		dst = *(type*)cpu->pc;			\
		cpu->pc += sizeof(type);		\
	} while (0)

#define pop_f()						\
	do {						\
		cpu->sp = (iptr*)((real*)cpu->sp - 1);	\
		ftmp = *(real*)cpu->sp;			\
	} while (0)
	
void do_syscall(struct cpu *cpu);

static inline void do_alloc(struct cpu *cpu)
{
	u32 class_id;

	get_unaligned(class_id, u32);
	
	cpu->this = malloc(cpu->d);
	memset(cpu->this, 0, cpu->d);
	*(u32*)cpu->this = class_id;
	
	if (class_id > (u32)cpu->nclasses)
		longjmp(cpu->hlt, OOPS_BADCLASS);
	if (cpu->sys == 0 && class_id == 0)
		cpu->sys = cpu->this;
}

static inline void do_new(struct cpu *cpu)
{
	struct run_class *self;
	i32 class_id;

	/* find out class id */
	get_unaligned(class_id, i32);

	/* do normal pre-call processing */
	*cpu->sp++ = (iptr) cpu->this;
	*cpu->sp++ = (iptr) cpu->pc;

	if (class_id > cpu->nclasses)
		longjmp(cpu->hlt, OOPS_BADCLASS);
		
	self = &cpu->classes[class_id];

	/* note that func[0] is constructor, it doesn't require
	 * %this to be set to anything intresting...
	 */
	cpu->pc = cpu->base_code + self->funcs[0];
}

static inline void do_jsr_id(struct cpu *cpu)
{
	struct run_class *self;
	int class_id;
	
	/* find out class id */
	get_unaligned(class_id, i32);

	/* do normal pre-call processing */
	*cpu->sp++ = (iptr) cpu->this;
	*cpu->sp++ = (iptr) cpu->pc;
	
	cpu->this = (byte*)cpu->a;
	
	if (class_id > cpu->nclasses)
		longjmp(cpu->hlt, OOPS_BADCLASS);
		
	self = &cpu->classes[class_id];
	
	if (cpu->d < 0 || cpu->d > self->nfuncs)
		longjmp(cpu->hlt, OOPS_BADFUNC);
		
	cpu->pc = cpu->base_code + self->funcs[cpu->d];
}

static inline void do_jsr(struct cpu *cpu)
{
	struct run_class *self;
	int class_id;
	
	*cpu->sp++ = (iptr) cpu->this;
	*cpu->sp++ = (iptr) cpu->pc;
	
	cpu->this = (byte*)cpu->a;
	class_id = *cpu->this;
	
	if (class_id > cpu->nclasses)
		longjmp(cpu->hlt, OOPS_BADCLASS);
		
	self = &cpu->classes[class_id];
	
	if (cpu->d < 0 || cpu->d > self->nfuncs)
		longjmp(cpu->hlt, OOPS_BADFUNC);
		
	cpu->pc = cpu->base_code + self->funcs[cpu->d];
}

static inline void do_frame(struct cpu *cpu)
{
	u32 size;
	
	get_unaligned(size, u32);
	*(cpu->sp++) = (iptr)cpu->bp;
	cpu->bp = cpu->sp;
	/* note, that clearing memory is only for local
	 * strings so that set_s won't free junk ptrs...
	 * possibly making local strings in compiler should
	 * be fixed, so that this is no longer needed: FIXME:
	 */
	memset(cpu->bp, 0, size);
	cpu->sp = (iptr*)((byte*)cpu->sp + size);

	/* after call and frame stack looks like this:
	 * arg..
	 * arg1		-4(bp)
	 * this		-3(bp)
	 * pc		-2(bp)
	 * saved-bp	-1(bp)
	 * localarg	(bp)
	 */
}


static inline void do_rts(struct cpu *cpu)
{
	cpu->sp = cpu->bp;
	cpu->bp = (iptr*)*--cpu->sp;
	cpu->pc = (byte*)*--cpu->sp;
	cpu->this = (byte*)*--cpu->sp;
}

inline void set_s(struct cpu *cpu, char **dst, const char *src)
{
	(void)cpu;
	if (*dst)
		free(*dst);
	if (src)
		*dst = strdup(src);
	else
		*dst = 0;
}

static inline void do_add_s(struct cpu *cpu)
{
	char *o, *r;
	int n;
	
	o = (char*)(*--cpu->sp);
	n = strlen(o);
	r = malloc(n + strlen(cpu->s) + 1);
	strcpy(r, o);
	strcpy(r + n, cpu->s);
	free(o);
	free(cpu->s);
	cpu->s = r;
}

static inline int do_test_s(struct cpu *cpu)
{
	char *o;
	int r;
	
	o = (char*)(*--cpu->sp);
	r = strcmp(o, cpu->s);
	free(o);

	return r;
}

static inline void check_sp(struct cpu *cpu)
{
	if (cpu->sp + 10 > cpu->top_stack)
		longjmp(cpu->hlt, OOPS_STACK);
}

extern char *op_codes[];
int debug;

void exec_xbytes(struct cpu *cpu, int n)
{
	iptr tmp;
	real ftmp;
	
	while (n--) {
		if (debug)
			printf("exec %5d: 0x%02x : %-25s [d = %d f = %.2f]\n", 
				cpu->pc - cpu->base_code,
				*(cpu->pc), op_codes[*(cpu->pc)],
				cpu->d, cpu->f);
		if (cpu->pc - cpu->base_code == brk) 
			printf("BREAK at %d\n", brk);
		
		switch (*(cpu->pc)++) {
		default:
		case op_fault:
			longjmp(cpu->hlt, OOPS_ILL);
			break;
		case op_hlt:
			longjmp(cpu->hlt, OOPS_HLT);
			break;
		case op_nop:
			break;

/* load const -> %d */
		case op_clr_d:
			cpu->d = 0;
			break;
		case op_mov_b:
			cpu->d = *(i8*)cpu->pc++;
			break;
		case op_mov_w:
			get_unaligned(cpu->d, i16);
			break;
		case op_mov_l:
			get_unaligned(cpu->d, i32);
			break;
#ifdef i64
		case op_mov_q:
			get_unaligned(cpu->d, i64);
			break;
#endif

/* load const -> %f */
		case op_clr_f:
			cpu->f = 0.0;
			break;
		case op_mov_f:
			get_unaligned(cpu->f, real);
			break;

/* load const -> %s */
		case op_mov_s:
			get_unaligned(tmp, i32);
			set_s(cpu, &cpu->s, (char*)cpu->pc + tmp);
			break;

/* load adr -> %a */
		case op_lda_bp:
			cpu->a = (iptr*)((byte*)cpu->bp + cpu->d);
			break;
		case op_lda_this:
			cpu->a = (iptr*)((byte*)cpu->this + cpu->d);
			break;
		case op_lda_sys:
			cpu->a = (iptr*)((byte*)cpu->sys + cpu->d);
			break;
		case op_add_a:
			cpu->a = (iptr*)((byte*)cpu->a + cpu->d);
			break;

/* load (%a) -> X */
		case op_mov_aa:
			cpu->a = (iptr*)*cpu->a;
			break;
		case op_mov_ad:
			cpu->d = *cpu->a;
			break;
		case op_mov_af:
			cpu->f = *(real*)cpu->a;
			break;
		case op_mov_as:
			set_s(cpu, &cpu->s, (char*)*cpu->a);
			break;
/* store X -> (%a) */
		case op_mov_da:
			*cpu->a = cpu->d;
			break;
		case op_mov_fa:
			*(real*)cpu->a = cpu->f;
			break;
		case op_mov_sa:
			set_s(cpu, (char**)cpu->a, cpu->s);
			break;

/* stack push */
		case op_push_d:
			*cpu->sp++ = cpu->d;
			check_sp(cpu);
			break;
		case op_push_a:
			*cpu->sp++ = (iptr)cpu->a;
			check_sp(cpu);
			break;
		case op_push_f:
			*((real*)cpu->sp) = cpu->f;
			cpu->sp = (iptr*)((real*)cpu->sp + 1);
			check_sp(cpu);
			break;
		case op_push_s:
			*(char**)(cpu->sp) = 0;
			set_s(cpu, (char**)(cpu->sp++), cpu->s);
			check_sp(cpu);
			break;

/* stack pop */
		case op_pop_d:
			cpu->d = *--cpu->sp;
			break;
		case op_pop_a:
			cpu->a = (iptr*)*--cpu->sp;
			break;
		case op_pop_f:
			cpu->sp = (iptr*)((real*)cpu->sp - 1);
			cpu->f = *((real*)cpu->sp);
			break;
		case op_pop_s:
			set_s(cpu, &cpu->s, (char*)*(--cpu->sp));
			set_s(cpu, (char**)cpu->sp, 0);
			break;

/* stack other: */
		case op_sub_sp:
			cpu->sp = (iptr*)((byte*)cpu->sp - cpu->d);
			break;

/* arithm */
/* reg only: */
		case op_neg_d:
			cpu->d = -cpu->d;
			break;
		case op_neg_f:
			cpu->f = -cpu->f;
			break;
		case op_inc_d:
			cpu->d++;
			break;
		case op_inc_f:
			cpu->f++;
			break;
		case op_dec_d:
			cpu->d--;
			break;
		case op_dec_f:
			cpu->f--;
			break;

/* reg + stack, REG = POP (op) REG */
		case op_add_d:
			cpu->d += *--cpu->sp;
			break;
		case op_add_f:
			pop_f();
			cpu->f += ftmp;
			break;
		case op_add_s:
			do_add_s(cpu);
			break;
		case op_sub_d:
			cpu->d = *--cpu->sp - cpu->d;
			break;
		case op_sub_f:
			pop_f();
			cpu->f = ftmp - cpu->f;
			break;
		case op_mul_d:
			cpu->d *= *--cpu->sp;
			break;
		case op_mul_f:
			pop_f();
			cpu->f *= ftmp;
			break;
		case op_div_d:
			if (cpu->d == 0)
				longjmp(cpu->hlt, OOPS_DIV0);
			cpu->d = *--cpu->sp / cpu->d;
			break;
		case op_div_f:
			pop_f();
			cpu->f = ftmp / cpu->f;
			break;
		case op_mod_d:
			if (cpu->d == 0)
				longjmp(cpu->hlt, OOPS_MOD0);
			cpu->d = *--cpu->sp % cpu->d;
			break;
		case op_and_d:
			cpu->d &= *--cpu->sp;
			break;
		case op_or_d:
			cpu->d |= *--cpu->sp;
			break;
		case op_eor_d:
			cpu->d ^= *--cpu->sp;
			break;
		case op_shl_d:
			cpu->d = *--cpu->sp << cpu->d;
			break;
		case op_shr_d:
			cpu->d = *--cpu->sp >> cpu->d;
			break;
/* logical: REG = POP (op) REG, note that to allow C-like execution of
 * if (a && b) jump code is also required ... */
		case op_not:
			cpu->d = !cpu->d;
			break;
		case op_and:
			cpu->d = *--cpu->sp && cpu->d;
			break;
		case op_or:
			cpu->d = *--cpu->sp || cpu->d;
			break;
		case op_eor:
			cpu->d = (*--cpu->sp & 1) ^ (cpu->d & 1);
			break;
			
/* converions, all is done in regs ! */
		case op_conv_fd:
			cpu->d = (int)(cpu->f + 0.5);
			break;
		case op_conv_df:
			cpu->f = cpu->d;
			break;

/* relations: */
		case op_teq_d:
			cpu->d = *--cpu->sp == cpu->d;
			break;
		case op_teq_s:
			cpu->d = do_test_s(cpu) == 0;
			break;
		case op_teq_f:
			pop_f();
			cpu->d = ftmp == cpu->f;
			break;

		case op_tne_d:
			cpu->d = *--cpu->sp != cpu->d;
			break;
		case op_tne_s:
			cpu->d = do_test_s(cpu) != 0;
			break;
		case op_tne_f:
			pop_f();
			cpu->d = ftmp != cpu->f;
			break;

		case op_tle_d:
			cpu->d = *--cpu->sp <= cpu->d;
			break;
		case op_tle_s:
			cpu->d = do_test_s(cpu) <= 0;
			break;
		case op_tle_f:
			pop_f();
			cpu->d = ftmp <= cpu->f;
			break;

		case op_tge_d:
			cpu->d = *--cpu->sp >= cpu->d;
			break;
		case op_tge_s:
			cpu->d = do_test_s(cpu) >= 0;
			break;
		case op_tge_f:
			pop_f();
			cpu->d = ftmp >= cpu->f;
			break;

		case op_tlt_d:
			cpu->d = *--cpu->sp <  cpu->d;
			break;
		case op_tlt_s:
			cpu->d = do_test_s(cpu) <  0;
			break;
		case op_tlt_f:
			pop_f();
			cpu->d = ftmp <  cpu->f;
			break;

		case op_tgt_d:
			cpu->d = *--cpu->sp >  cpu->d;
			break;
		case op_tgt_s:
			cpu->d = do_test_s(cpu) >  0;
			break;
		case op_tgt_f:
			pop_f();
			cpu->d = ftmp >  cpu->f;
			break;
/* relative jumps */
		case op_bra:
			get_unaligned(tmp, i32);
			cpu->pc += tmp;
			break;
		case op_btrue:
			get_unaligned(tmp, i32);
			if (cpu->d)
				cpu->pc += tmp;
			break;
		case op_bfalse:
			get_unaligned(tmp, i32);
			if (!cpu->d)
				cpu->pc += tmp;
			break;

/* calling ... */
		case op_syscall:
			do_syscall(cpu);
			break;
			
		case op_alloc:
			do_alloc(cpu);
			break;

		case op_die:
			free(cpu->this);
			break;

		case op_new:
			do_new(cpu);
			break;

		case op_jsr_id:
			do_jsr_id(cpu);
			break;
			
		case op_jsr:
			do_jsr(cpu);
			break;
			
		case op_frame:
			do_frame(cpu);
			break;
			
		case op_rts:
			do_rts(cpu);
			break;

		case op_swap_d:
			tmp = cpu->d;
			cpu->d = *--cpu->sp;
			*cpu->sp++ = tmp;
			break;

		case op_swap_f:
			pop_f();
			*((real*)cpu->sp) = cpu->f;
			cpu->sp = (iptr*)((real*)cpu->sp + 1);
			cpu->f = ftmp;
			break;
			
		case op_swap_s:
			tmp = (iptr)cpu->s;
			cpu->s = (char*)*--cpu->sp;
			*cpu->sp++ = tmp;
			break;

		case op_test_f:
			cpu->d = cpu->f != 0;
			break;

		case op_mov_dA:
			cpu->a = (iptr*)cpu->d;
			break;

		case op_mov_Ad:
			cpu->d = (iptr)cpu->a;
			break;
		
		case op_mov_this_a:
			cpu->a = (iptr*)cpu->this;
			break;

		case op_mov_sys_a:
			cpu->a = (iptr*)cpu->sys;
			break;
			
		case op_save_ds:
			cpu->save_s = cpu->s;
			cpu->s = 0;
			cpu->save_d = cpu->d;
			break;
			
		case op_load_ds:
			cpu->s = cpu->save_s;
			cpu->d = cpu->save_d;
			break;

		}
	}
}

struct cpu *load(const char *fn)
{
	struct cpu *cpu;
	struct kbin_file *o;
	struct kbin_tag *t;
	byte *code = 0;
	struct run_class *v = 0;
	int nc = 0, i;
	u32 *ptr;
	
	o = kbin_open(fn, 1, 0);
	if (o == 0)
		return 0;

	while ((t = kbin_get_tag(o))) {
		if (t->id == kbin_tag_eof) {
			kbin_kill_tag(t);
			break;
		}

		switch (t->id) {
		case kbin_tag_code:
			code = t->data;
			t->data = 0;
			break;
		case kbin_tag_vtables:
			ptr = (u32*)t->data;
			t->data = 0;
			nc = *ptr++;
			v = (struct run_class*)malloc(
				sizeof(struct run_class) * nc);
			for (i = 0; i < nc; i++) {
				v[i].nfuncs = *ptr++;
				v[i].funcs = (i32*)ptr;
				ptr += v[i].nfuncs;
			}
			break;
		}

		kbin_kill_tag(t);
	}

	kbin_close(o);
	
	if (t == 0 || code == 0 || v == 0)
		return 0;
		
	cpu = NEW(struct cpu);
	memset(cpu, 0, sizeof(*cpu));
	
	cpu->sp = (iptr*)malloc(sizeof(iptr) * STACK_SIZE);
	cpu->top_stack = cpu->sp + STACK_SIZE;
	cpu->pc = cpu->base_code = code;

	cpu->classes = v;
	cpu->nclasses = nc;

	return cpu;
}

static int handle_oops(struct cpu *cpu, int n)
{
	char *oops[] = {
		"XXX",
		"OOPS_ILL",
		"OOPS_HLT",
		"OOPS_STACK",
		"OOPS_DIV0",
		"OOPS_MOD0",
		"OOPS_BADFUNC",
		"OOPS_BADCLASS",
		"OOPS_BADSYSCALL",
		"OOPS_EXIT",
	};
	char *e;

	if (n > OOPS_EXIT)
		e = "BAD OOPS";
	else if (n == OOPS_EXIT)
		e = 0;
	else
		e = oops[n];

	if (e == 0)
		exit(cpu->d);
	
	fprintf(stderr, "%s\n", e);
	exit(1);
}

int main(int argc, char **argv)
{
	struct cpu *cpu;
	char *name, *p;
	int r;
	
	(void)argc;
	name = *argv++;
	while (*argv && **argv == '-') {
		for (p = argv[0] + 1; *p; p++)
			switch (*p) {
			case 'd':
				debug = 1;
				break;
			default:
				fprintf(stderr, "Unknown option: -%c\n", *p);
				exit(1);
			}
		argv++;
	}
	
	if (*argv == 0) {
		fprintf(stderr, "USAGE: %s [-d] filename\n", name);
		exit(1);
	}
	
	cpu = load(*argv);
	r = setjmp(cpu->hlt);
	if (r != 0)
		handle_oops(cpu, r);

	for (;;)
		exec_xbytes(cpu, 1024);
}
